/*****************************************************************************
 *
 * Europa Firmware Update Utility
 * Copyright 2000-2001, Synthcom Systems, Inc.
 *
 * This module handles all the MIDI control and timing
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "hosttype.h"
#include "europa.h"
#include "efile.h"
#include "emidi.h"
#include "host.h"

#define	MAX_MIDI_RX		50		// 50  Bytes Response maximum!

// A couple of standard MIDI bytes

#define	SYSEX_START		0xf0
#define SYSEX_END		0xf7

static const UINT8 sg_u8EuropaSysexHeader[] = 
{
	SYSEX_START,
	0x41,
	0x10,
	0x4a,
	0x36
};

struct SProgramToEuropaErrorXlate
{
	EEuropaProgramResponses eProgramResponse;
	EEuropaError eEuropaError;	
};

static const struct SProgramToEuropaErrorXlate sg_sErrorTable[] =
{
	{PROGRAM_BAD_CHECKSUM,				EUROPA_BAD_CHECKSUM},
	{PROGRAM_BAD_LENGTH,				EUROPA_BAD_LENGTH},
	{PROGRAM_BAD_BLOCK_SEQUENCE,		EUROPA_BAD_SEQUENCE},
	{PROGRAM_UNEXPECTED,				EUROPA_BAD_RESPONSE},
	{PROGRAM_FAULT,						EUROPA_PROGRAM_FAULT},
	{PROGRAM_COMPLETE_BAD_CHECKCODE,	EUROPA_BAD_CHECKCODE},
	{PROGRAM_DATA_TOO_LONG,				EUROPA_BAD_LENGTH}
};

;// Various states of the receive state machine

typedef enum
{
	RSTATE_WAIT_SYSEX,
	RSTATE_WAIT_41,
	RSTATE_WAIT_ANY,
	RSTATE_WAIT_4a,
	RSTATE_WAIT_36,
	RSTATE_WAIT_BYTE
} ERXState;

static volatile BOOL sg_bMIDICanceled = FALSE;
static volatile UINT8 sg_u8ReceiveBuffer[MAX_MIDI_RX];
static volatile UINT32 sg_u32ReceiveHead = 0;
static volatile UINT32 sg_u32ReceiveTail = 0;
static ERXState sg_eRXState = RSTATE_WAIT_SYSEX;

/*****************************************************************************
 *
 * Procedure  : EuropaResetBoard
 *
 * Input      : Nothing
 *
 * Output     : Nothing
 *
 * Description:
 *
 * This routine will cancel any waiting for MIDI (if any)
 *
 *****************************************************************************/

void EuropaMIDICancel(void)
{
	sg_bMIDICanceled = TRUE;
}

/*****************************************************************************
 *
 * Procedure  : EuropaReceiveMIDIByte
 *
 * Input      : MIDI Byte received by the platform
 *
 * Output     : Nothing
 *
 * Description:
 *
 * This routine is called when a MIDI byte is received. This receive state machine
 * will filter out anything not bound for Europa and will not place anything in
 * the receive buffer 
 *
 *****************************************************************************/

void EuropaReceiveMIDIByte(UINT8 u8RXByte)
{
	// If we receive any byte that's not a sysex control byte but it's another
	// control message, reset our state machine so we can resync

checkReceivedByte:

	if ((u8RXByte >= 0x80) && (u8RXByte != SYSEX_START) && (u8RXByte != SYSEX_END))
	{
		sg_eRXState = RSTATE_WAIT_SYSEX;
	}

	// If we're waiting for a Sysex, only a sysex will do

	if (RSTATE_WAIT_SYSEX == sg_eRXState)
	{
		if (u8RXByte == SYSEX_START)
		{
			sg_eRXState = RSTATE_WAIT_41;
		}
	}
	else
	if (RSTATE_WAIT_41 == sg_eRXState)
	{
		if (u8RXByte != 0x41)
		{
			sg_eRXState = RSTATE_WAIT_SYSEX;
			goto checkReceivedByte;
		}

		sg_eRXState = RSTATE_WAIT_ANY;
	}
	else
	if (RSTATE_WAIT_ANY == sg_eRXState)
	{
		sg_eRXState = RSTATE_WAIT_4a;
	}
	else
	if (RSTATE_WAIT_4a == sg_eRXState)
	{
		if (u8RXByte != 0x4a)
		{
			sg_eRXState = RSTATE_WAIT_SYSEX;
			goto checkReceivedByte;
		}

		sg_eRXState = RSTATE_WAIT_36;
	}
	else
	if (RSTATE_WAIT_36 == sg_eRXState)
	{
		if (u8RXByte != 0x36)
		{
			sg_eRXState = RSTATE_WAIT_SYSEX;
			goto checkReceivedByte;
		}

		sg_eRXState = RSTATE_WAIT_BYTE;
	}
	else	// We're receiving a data byte
	{
		sg_u8ReceiveBuffer[sg_u32ReceiveHead++] = u8RXByte;

		// Circular buffer - we need to wrap

		if (sg_u32ReceiveHead >= sizeof(sg_u8ReceiveBuffer))
		{
			sg_u32ReceiveHead = 0;
		}

		// If we've hit the beginning of the buffer, we need to wrap

		if (sg_u32ReceiveHead == sg_u32ReceiveTail)
		{
			sg_u32ReceiveTail++;
			if (sg_u32ReceiveTail >= sizeof(sg_u8ReceiveBuffer))
			{
				sg_u32ReceiveTail = 0;
			}
		}

		// If we get a sysex receive end, make sure the sysex end is plunked
		// in the receive circular buffer and reset the state machine.

		if (SYSEX_END == u8RXByte)
		{
			sg_eRXState = RSTATE_WAIT_SYSEX;
		}
	}
}

/*****************************************************************************
 *
 * Procedure  : EuropaFlushMIDIStream
 *
 * Input      : Nothing
 *
 * Output     : Nothing
 *
 * Description:
 *
 * This routine is called to eliminate any "junk" in the MIDI stream on startup.
 *
 *****************************************************************************/

void EuropaFlushMIDIStream(void)
{
	sg_eRXState = RSTATE_WAIT_SYSEX;
	sg_u32ReceiveTail = 0;
	sg_u32ReceiveHead = 0;
}

/*****************************************************************************
 *
 * Procedure  : EuropaMIDIResetCmd
 *
 * Input      : Reset reason
 *
 * Output     : Nothing
 *
 * Description:
 *
 * This routine is called when Europa needs to be reset
 *
 *****************************************************************************/

void EuropaMIDIResetCmd(UINT8 u8ResetReason)
{
	UINT8 u8Message[sizeof(sg_u8EuropaSysexHeader) + 3];

	memcpy(u8Message, sg_u8EuropaSysexHeader, sizeof(sg_u8EuropaSysexHeader));
	u8Message[sizeof(sg_u8EuropaSysexHeader)] = ESYSEX_RESET;
	u8Message[sizeof(sg_u8EuropaSysexHeader) + 1] = u8ResetReason;
	u8Message[sizeof(sg_u8EuropaSysexHeader) + 2] = SYSEX_END;

	// Go send the MIDI block!

	HostSendMIDIBlock(u8Message, sizeof(u8Message));
}

/*****************************************************************************
 *
 * Procedure  : EuropaMIDISendGetStatusRequest
 *
 * Input      : Nothing
 *
 * Output     : Nothing
 *
 * Description:
 *
 * This routine is called when a status info block command is sent
 *
 *****************************************************************************/

void EuropaMIDISendGetStatusRequest(void)
{
	UINT8 u8Message[sizeof(sg_u8EuropaSysexHeader) + 2];

	memcpy(u8Message, sg_u8EuropaSysexHeader, sizeof(sg_u8EuropaSysexHeader));
	u8Message[sizeof(sg_u8EuropaSysexHeader)] = ESYSEX_GET_REVS;
	u8Message[sizeof(sg_u8EuropaSysexHeader) + 1] = SYSEX_END;

	// Go send the MIDI block!

	HostSendMIDIBlock(u8Message, sizeof(u8Message));
}

/*****************************************************************************
 *
 * Procedure  : EuropaMIDISendPrepareArea
 *
 * Input      : Nothing
 *
 * Output     : Nothing
 *
 * Description:
 *
 * This routine is called when a status info block command is sent
 *
 *****************************************************************************/

extern void EuropaMIDISendPrepareArea(EEuropaFileCommand eCmd)
{
	UINT8 u8Message[sizeof(sg_u8EuropaSysexHeader) + 3];

	memcpy(u8Message, sg_u8EuropaSysexHeader, sizeof(sg_u8EuropaSysexHeader));
	u8Message[sizeof(sg_u8EuropaSysexHeader)] = ESYSEX_PREPARE_AREA;
	u8Message[sizeof(sg_u8EuropaSysexHeader) + 1] = eCmd;
	u8Message[sizeof(sg_u8EuropaSysexHeader) + 2] = SYSEX_END;

	// Go send the MIDI block!

	HostSendMIDIBlock(u8Message, sizeof(u8Message));
}

/*****************************************************************************
 *
 * Procedure  : EuropaMIDIGetResponse
 *
 * Input      : Pointer to buffer where response goes 
 *				Response buffer size
 *				# Of tenths of seconds to wait for a Europa sysex stream
 *
 * Output     : Nothing
 *
 * Description:
 *
 * This routine is called when a status info block command is sent
 *
 *****************************************************************************/

EEuropaError EuropaMIDIGetResponse(UINT8 *pu8ResponseBuffer,
								   UINT32 u32ResponseBufferSize,
								   UINT32 *pu32ResponseSize,
								   UINT32 u32TenthsOfSeconds)
{
	UINT8 u8DataByte;

	sg_bMIDICanceled = FALSE;		// Indicate we've not received anything

	memset(pu8ResponseBuffer, 0, u32ResponseBufferSize);

	// If they've passed us a pointer to a response size count, zero it out

	if (pu32ResponseSize)
	{
		*pu32ResponseSize = 0;
	}

	// Loop through our tenths of seconds or until we've received a full
	// MIDI response message

	while ((u32TenthsOfSeconds--) && (FALSE == sg_bCanceled))
	{
		// If our head and tail pointers are equal then we should delay to
		// give time for things to be received

		if (sg_u32ReceiveTail == sg_u32ReceiveHead)
		{
			HostSleep(1);		// Sleep 100ms
		}
		else
		{
			u8DataByte = sg_u8ReceiveBuffer[sg_u32ReceiveTail++];

			if (sg_u32ReceiveTail >= sizeof(sg_u8ReceiveBuffer))
			{
				sg_u32ReceiveTail = 0;
			}

			// If this is a system exclusive ending, let's return our response
			// as a EUROPA_NO_ERROR
	
			if (SYSEX_END == u8DataByte)
			{
				return(EUROPA_NO_ERROR);
			}

			// Let's see if we're OK or if things are screwed up

			*pu8ResponseBuffer = u8DataByte;
			++pu8ResponseBuffer;
			u32ResponseBufferSize--;		

			// If there's really a response size, increment its count

			if (pu32ResponseSize)
			{
				++(*pu32ResponseSize);
			}

			// If our response buffer size == 0, then we're out of room and
			// should indicate so

			if (0 == u32ResponseBufferSize)
			{
				return(EUROPA_RESPONSE_TOO_LONG);
			}
		}
	}

	if (sg_bCanceled)
	{
		return(EUROPA_CANCELED);
	}

	return(EUROPA_RESPONSE_TIMEOUT);
}

/*****************************************************************************
 *
 * Procedure  : EuropaMIDIProgramBlock
 *
 * Input      : Block # to program, pointer to data block to program, and the
 *				length of the block to program.
 *
 * Output     : Europa error (see europa.h)
 *
 * Description:
 *
 * This routine will program a block of data to Europa and wait for an
 * appropriate response.
 *
 *****************************************************************************/

EEuropaError EuropaMIDIProgramBlock(UINT32 u32BlockNumber,
									UINT8 *pu8DataArea,
									UINT32 u32Length)
{
	EEuropaError eError = EUROPA_NO_ERROR;
	UINT32 u32Checksum;
	UINT32 u32Loop;
	UINT32 u32ResponseLength;
	UINT8 u8ResponseBuffer[50];
	UINT8 *pu8DataBlock = NULL;
	UINT8 *pu8DataHead;
	UINT8 *pu8Data;

	// First things first - allocate some memory for the transfer. Make sure
	// to add in the sysex terminator, the block #, the check code, the command,
	// and the data block length.

	pu8DataBlock = malloc(u32Length + sizeof(sg_u8EuropaSysexHeader) + 8);

	// If we didn't get the block, indicate an error!

	if (NULL == pu8DataBlock)
	{
		eError = EUROPA_NO_MEMORY;
		goto programBlockExit;
	}

	memset(pu8DataBlock, 0, u32Length + sizeof(sg_u8EuropaSysexHeader) + 8);

	// Now copy all pertinent information in place and get it ready for
	// transfer

	memcpy(pu8DataBlock, sg_u8EuropaSysexHeader, sizeof(sg_u8EuropaSysexHeader));
	pu8Data = pu8DataBlock + sizeof(sg_u8EuropaSysexHeader);

	// Give the programming command

	*pu8Data++ = (UINT8) ESYSEX_PROGRAM_BLOCK;

	// Block number - MSB (7 bits, actually) first

	u32Checksum = 0;
	pu8DataHead = pu8Data;

	*pu8Data++ = (UINT8) ((u32BlockNumber >> 7) & 0x7f);
	*pu8Data++ = (UINT8) (u32BlockNumber & 0x7f);

	// Length of block - MSB (7 bits) first

	*pu8Data++ = (UINT8) ((u32Length >> 7) & 0x7f);
	*pu8Data++ = (UINT8) (u32Length & 0x7f);

	// Now let's copy in the data!

	memcpy(pu8Data, pu8DataArea, u32Length);

	for (u32Loop = 0; u32Loop < (u32Length + 4); u32Loop++)
	{
		u32Checksum += *pu8DataHead;
		++pu8DataHead;
	}

	pu8Data += u32Length;

	// Now tuck away the checksum

	*pu8Data++ = (UINT8) ((u32Checksum >> 7) & 0x7f);
	*pu8Data++ = (UINT8) (u32Checksum & 0x7f);
	
	// Terminate it with an end sysex message

	*pu8Data++ = SYSEX_END;

	// We've got our full message. Let's send it to the platform specific
	// code so it'll get transmitted

	HostSendMIDIBlock(pu8DataBlock, (UINT32) ((UINT32) pu8Data - (UINT32) pu8DataBlock));
	
	// We've sent it! Now let's wait for a response - up to 10 seconds for it to
	// program.

	eError = EuropaMIDIGetResponse(u8ResponseBuffer, sizeof(u8ResponseBuffer), &u32ResponseLength, 10*10);

	// If we don't have an error, we've got a response. Let's make sure it's
	// completely and absolutely valid.
	
	if (EUROPA_NO_ERROR == eError)
	{
		// Make sure it's the response we expected
	
		if (ESYSEX_PROGRAM_RESPONSE == u8ResponseBuffer[0])
		{
			// It's a program response command - let's see if it's a non-OK message. If
			// so, we need to look it up and translate it to a generic Europa error msg.

			if ((PROGRAM_COMPLETE_IMAGE_OK == u8ResponseBuffer[1]) && (2 == u32ResponseLength))
			{
				// If we have a completed image and it's OK and there's a 0 length block
				// to be programmed, just return that it's OK
			}
			else
			if (PROGRAM_OK == u8ResponseBuffer[1])
			{
				// Program is OK!
			}
			else
			{
				// Here we know we've got a bad response. Look up the program response
				// and translate it to a generic Europa error. If not found, indicate
				// EUROPA_BAD_RESPONSE
	
				eError = EUROPA_BAD_RESPONSE;

				for (u32Loop = 0; u32Loop < (sizeof(sg_sErrorTable) / sizeof(sg_sErrorTable[0])); u32Loop++)
				{
					if (sg_sErrorTable[0].eProgramResponse == (EEuropaProgramResponses) u8ResponseBuffer[1])
					{
						eError = sg_sErrorTable[0].eEuropaError;
						break;
					}
				}
			}
		}
		else
		{
			eError = EUROPA_BAD_RESPONSE;	// Bogus response
		}
	}

programBlockExit:

	// If we've allocated a block, deallocate it!

	if (pu8DataBlock)
	{
		free(pu8DataBlock);
	}

	return(eError);
}

